// Traveling Salesman Problem via SetBuilder Algorithm
// Written by David J. Stein, Esq. See "License.html" for terms and conditions of use.
//
// This module contains the code for the SetBuilder class.

#region Using declarations

using System;
using System.IO;
using System.Collections;

#endregion

namespace Set_Builder_Algorithm_Namespace {

	public class SetBuilderAlgorithm {

		#region Variables and initialization code

		public ArrayList cities;
		public int[,] iDistances;
		public int iMaxSize;
		public Random r;
		public ArrayList[] lists;
		public int iMaxDistance;
		public decimal dThreshold;
		public int[] iListMaxSize;

		public SetBuilderAlgorithm() {
			r = new Random((int) System.DateTime.Now.Ticks);
			iMaxSize = 500;
			dThreshold = 0.5M;
			cities = new ArrayList();
			Initialize();
		}

		#endregion

		#region Top-level (core) function

		public void Solve() {
			// This function creates a new solution and attempts to add it to the pile.
			char i = (char) r.Next(cities.Count);
			char j = (char) r.Next(cities.Count);
			while (j == i)
				j = (char) r.Next(cities.Count);
			Entity e = new Entity(cities.Count);
			e.components.Add(i);
			e.components.Add(j);
			e.length = 2;
			e.path = Assemble(e);
			e.signature.endpoints[0] = i;
			e.signature.endpoints[1] = j;
			e.signature.Add(i);
			e.signature.Add(j);
			SetFitness(e);
			Insert(e);

			// generate upper-level entities
			for (int level = 1; level < lists.Length; level++) {
				if (lists[level - 1].Count < iListMaxSize[level - 1] * dThreshold)  // reached top of stack - exit
					break;
				Entity newEntity = new Entity(cities.Count);
				for (int trials = 0; trials < 100 & newEntity.length < level + 2; trials++) {
					if (newEntity.length == level + 1) {  // only one left - choose and insert randomly
						char newCity = (char) r.Next(cities.Count);
						while (newEntity.signature.CheckOverlap(newCity) == true)
							newCity = (char) r.Next(cities.Count);
						ArrayList originalComponents = (ArrayList) newEntity.components.Clone();
						decimal best = 0;
						int bestPosition = 0;
						for (int iPosition = 0; iPosition < newEntity.components.Count + 1; iPosition++) {
							newEntity.components.Insert(iPosition, newCity);
							newEntity.path = Assemble(newEntity);
							SetFitness(e);
							if ((best == 0) || (best < newEntity.fitness)) {
								bestPosition = iPosition;
								best = newEntity.fitness;
							}
							newEntity.components.RemoveAt(iPosition);
						}
						newEntity.components = originalComponents;
						newEntity.components.Insert(bestPosition, newCity);
						newEntity.length++;
						newEntity.path = Assemble(newEntity);
						newEntity.signature.Add(newCity);
						if (bestPosition == 0)
							newEntity.signature.endpoints[0] = newCity;
						else if (bestPosition == newEntity.components.Count)
							newEntity.signature.endpoints[1] = newCity;
						SetFitness(newEntity);
					}
					else {  // choose a new starting piece
						// choose starting piece
						int top = Math.Max(level - (newEntity.length + 2), 0);
						int bottom = Math.Max((top - 1) / 2, 0);
						int iStartLevel = top;
						if (bottom != top)
							iStartLevel = r.Next(top - bottom) + bottom;
						Entity addEntity = (Entity) lists[iStartLevel][r.Next(lists[iStartLevel].Count)];
						if (newEntity.signature.CheckOverlap(addEntity.signature) == false) {  // add
							newEntity.components.Insert(newEntity.components.Count, addEntity);
							newEntity.length += addEntity.length;
							newEntity.path = Assemble(newEntity);
							newEntity.signature.Add(addEntity.signature);
							newEntity.signature.endpoints[1] = addEntity.signature.endpoints[1];
							SetFitness(newEntity);
						}
					}
				}
				if (newEntity.length == level + 2)   // successful creation
					Insert(newEntity);
			}
		}

		#endregion

		#region Mid-level (attendant) functions

		public void Initialize() {
			// This function initializes the pile.
			if (cities.Count < 2)
				lists = null;
			else {
				lists = new ArrayList[cities.Count - 1];
				iListMaxSize = new int[cities.Count - 1];
				for (int i = 0; i < cities.Count - 1; i++) {
					lists[i] = new ArrayList();
					iListMaxSize[i] = (int) Math.Max(1, (iMaxSize * (1.0 - ((float) i ) / (float) (cities.Count - 2))));
				}
			}
		}

		public void SetFitness(Entity e) {
			// This function computes the fitness of an entity.
			if (e.length < 2) {
				e.fitness = e.distance = 0;
			}
			else {
				e.distance = CalculateEntityDistance(e);
				e.fitness = -e.distance - (iDistances[e.path[0], e.path[e.length - 1]] * (e.path.Length / cities.Count));
			}
		}

		public bool Insert(Entity e) {
			// This function attempts to insert an entity into the pile, and also (if necessary) reoptimizing the list and deleting (and detaching) any outcompeted entities.
			if ((e == null) || (e.length < 2))
				return false;
			ArrayList list = lists[e.length - 2];
			int iPosition = 0;
			for (; iPosition < list.Count; iPosition++) {
				Entity e2 = (Entity) list[iPosition];
				if (e2.fitness < e.fitness)
					break;
				else if (e.signature.CheckEquivalence(e2.signature) == true)
					return false;  // found another string with better signature
			}
			// found spot
			list.Insert(iPosition, e);
			// add children
			foreach (object o in e.components) {
				if (o is Entity)
					((Entity) o).references.Add(e);
			}
			// compare lists after this one to see if we've improved this position
			for (++iPosition; iPosition < list.Count; iPosition++) {
				Entity e2 = (Entity) list[iPosition];
				if (e.signature.CheckEquivalence(e2.signature) == true) {  // found signature
					e.references = e2.references;
					foreach (Entity parent in e.references)
						Replace(parent, e, e2);
					e2.references = new ArrayList();
					Remove(e2);
					return true;
				}
			}
			while (list.Count > iListMaxSize[e.length - 2])
				Remove((Entity) list[list.Count - 1]);
			return true;
		}

		public void Replace(Entity parent, Entity newChild, Entity oldChild) {
			// This function replaces a child with an improved child and optimizes all parents.
			for (int i = 0; i < parent.components.Count; i++) {
				if (parent.components[i] == oldChild) {
					parent.components[i] = newChild;
					break;
				}
			}
			AdjustFitness(parent);
		}

		public void AdjustFitness(Entity parent) {
			// This function reorders an improved parent in the pile. 
			parent.path = Assemble(parent);
			SetFitness(parent);
			// resort list
			ArrayList list = lists[parent.length - 2];
			int index = list.IndexOf(parent);
			if ((index > 0) && (((Entity) list[index - 1]).fitness < parent.fitness)) {  // must re-sort list
				int i = index - 2;
				for (; i >= 0; i--) {
					if (((Entity) list[i]).fitness > parent.fitness) {  // insert here
						list.RemoveAt(index);
						list.Insert(i + 1, parent);
						break;
					}
				}
				if (i == -1) {  // move to the head of the class!
					list.RemoveAt(index);
					list.Insert(0, parent);
				}
			}
			// adjust all parents
			foreach (Entity superparent in parent.references)
				AdjustFitness(superparent);
		}

		#endregion

		#region Low-level (helper) functions

		public string Assemble(Entity e) {
			// This function assembles the string path for an entity.
			string s = "";
			foreach (object o in e.components) {
				if (o is Entity)
					s += ((Entity) o).path;
				else
					s += (char) o;
			}
			return s;
		}

		public decimal CalculateEntityDistance(Entity e) {
			// This function calculates the total distance in the solution of an entity.
			decimal iTotalDistance = 0;
			for (int i = 1; i < e.length; i++)
				iTotalDistance += iDistances[e.path[i - 1], e.path[i]];
			return iTotalDistance;
		}

		public void Remove(Entity e) {
			// This function removes an entity from the pile, and detaches it from all parents in the pile.
			// remove from list
			lists[e.length - 2].Remove(e);
			// decrement all children
			foreach (object o in e.components) {
				if (o is Entity)
					RemoveParent((Entity) o, e);
			}
			// delete all parents
			while (e.references.Count > 0)
				Remove((Entity) e.references[0]);
		}

		public void RemoveParent(Entity child, Entity parent) {
			// This function detaches a child from a parent in the pile.
			child.references.Remove(parent);
		}

		public bool CompareStrings(string s1, string s2) {
			// This function compares two string paths.
			if ((s1 == null) || (s2 == null) || (s1.Length != s2.Length))
				return false;
			for (int i = 0; i < s1.Length; i++) {
				if (s1[i] != s2[i])
					return false;
			}
			return true;
		}

		#endregion

		#region City map functions

		public void Clear() {
			// This function erases all cities in the city list.
			this.cities = new ArrayList();
		}

		public void AddCity(int x, int y) {
			// This function creates a random city and adds it to the city list, and recalculates all city distances.
			int i = r.Next(138);
			string name = "";
			switch (i) {
				case 1: name = "Baltimore"; break;
				case 2: name = "Chicago"; break;
				case 3: name = "Cincinnati"; break;
				case 4: name = "New York City"; break;
				case 5: name = "Kansas City"; break;
				case 6: name = "San Diego"; break;
				case 7: name = "San Francisco"; break;
				case 8: name = "Paris"; break;
				case 9: name = "Berlin"; break;
				case 10: name = "Washington D.C."; break;
				case 11: name = "Cincinnati"; break;
				case 12: name = "Columbus"; break;
				case 13: name = "Philadelphia"; break;
				case 14: name = "New Orleans"; break;
				case 15: name = "Dallas"; break;
				case 16: name = "Salt Lake City"; break;
				case 17: name = "Houston"; break;
				case 18: name = "Los Angeles"; break;
				case 19: name = "Boston"; break;
				case 20: name = "Detroit"; break;
				case 21: name = "Toronto"; break;
				case 22: name = "Albany"; break;
				case 23: name = "Orlando"; break;
				case 24: name = "Anchorage"; break;
				case 25: name = "Brussels"; break;
				case 26: name = "London"; break;
				case 27: name = "Denver"; break;
				case 28: name = "San Antonio"; break;
				case 29: name = "Tampa"; break;
				case 30: name = "Miami"; break;
				case 31: name = "Atlanta"; break;
				case 32: name = "Providence"; break;
				case 33: name = "Tallahassee"; break;
				case 34: name = "Nashville"; break;
				case 35: name = "Las Vegas"; break;
				case 36: name = "Santa Fe"; break;
				case 37: name = "Boise"; break;
				case 38: name = "Seattle"; break;
				case 39: name = "Portland"; break;
				case 40: name = "Bismarck"; break;
				case 41: name = "Minneapolis"; break;
				case 42: name = "Milwaukee"; break;
				case 43: name = "Montreal"; break;
				case 44: name = "Winnipeg"; break;
				case 45: name = "Calgary"; break;
				case 46: name = "Birmingham"; break;
				case 47: name = "St. Louis"; break;
				case 48: name = "Little Rock"; break;
				case 49: name = "Oklahoma City"; break;
				case 50: name = "Indianapolis"; break;
				case 51: name = "Pittsburgh"; break;
				case 52: name = "Phoenix"; break;
				case 53: name = "Omaha"; break;
				case 54: name = "Wichita"; break;
				case 55: name = "Lexington"; break;
				case 56: name = "Louisville"; break;
				case 57: name = "Richmond"; break;
				case 58: name = "Charlotte"; break;
				case 59: name = "Savannah"; break;
				case 60: name = "Stalingrad"; break;
				case 61: name = "Leningrad"; break;
				case 62: name = "Moscow"; break;
				case 63: name = "Baghdad"; break;
				case 64: name = "Grand Rapids"; break;
				case 65: name = "Ottawa"; break;
				case 66: name = "Guatemala City"; break;
				case 67: name = "Mexico City"; break;
				case 68: name = "Havana"; break;
				case 69: name = "San Juan"; break;
				case 70: name = "Rio de Janeiro"; break;
				case 71: name = "Madrid"; break;
				case 72: name = "Barcelona"; break;
				case 73: name = "Marseilles"; break;
				case 74: name = "Amsterdam"; break;
				case 75: name = "Dublin"; break;
				case 76: name = "Glasgow"; break;
				case 77: name = "Frankfurt"; break;
				case 78: name = "Hamburg"; break;
				case 79: name = "Munich"; break;
				case 80: name = "Milan"; break;
				case 81: name = "Rome"; break;
				case 82: name = "Naples"; break;
				case 83: name = "Genoa"; break;
				case 84: name = "Belgrade"; break;
				case 85: name = "Zagreb"; break;
				case 86: name = "Sofia"; break;
				case 87: name = "Athens"; break;
				case 88: name = "Istanbul"; break;
				case 89: name = "Budapest"; break;
				case 90: name = "Warsaw"; break;
				case 91: name = "Kiev"; break;
				case 92: name = "Belarus"; break;
				case 93: name = "Stockholm"; break;
				case 94: name = "Oslo"; break;
				case 95: name = "Minsk"; break;
				case 96: name = "Kolkata"; break;
				case 97: name = "Beijing"; break;
				case 98: name = "Hong Kong"; break;
				case 99: name = "Tokyo"; break;
				case 100: name = "Mumbai"; break;
				case 101: name = "Bhopal"; break;
				case 102: name = "Delhi"; break;
				case 103: name = "Kabul"; break;
				case 104: name = "Mosul"; break;
				case 105: name = "Jerusalem"; break;
				case 106: name = "Damascus"; break;
				case 107: name = "Riyadh"; break;
				case 108: name = "Mecca"; break;
				case 109: name = "Dubai"; break;
				case 110: name = "Alexandria"; break;
				case 111: name = "Cairo"; break;
				case 112: name = "Tunis"; break;
				case 113: name = "Tripoli"; break;
				case 114: name = "Algiers"; break;
				case 115: name = "Beirut"; break;
				case 116: name = "Bangkok"; break;
				case 117: name = "Ho Chi Minh City"; break;
				case 118: name = "Hanoi"; break;
				case 119: name = "Chengdu"; break;
				case 120: name = "Guangzhou"; break;
				case 121: name = "Seoul"; break;
				case 122: name = "Shanghai"; break;
				case 123: name = "Phnom Penh"; break;
				case 124: name = "Mandalay"; break;
				case 125: name = "Kuala Lumpur"; break;
				case 126: name = "Singapore"; break;
				case 127: name = "Jakarta"; break;
				case 128: name = "Perth"; break;
				case 129: name = "Melbourne"; break;
				case 130: name = "Sydney"; break;
				case 131: name = "Brisbane"; break;
				case 132: name = "Hokkaido"; break;
				case 133: name = "Kyoto"; break;
				case 134: name = "Kagashima"; break;
				case 135: name = "Fukuoka"; break;
				case 136: name = "Murmansk"; break;
				case 137: name = "Turin"; break;
				case 138: name = "Lisbon"; break;
				default: name = "Cleveland"; break;
			}
			cities.Add(new City(x, y, name));
			CalculateDistances();
			lists = new ArrayList[cities.Count];
			Initialize();
		}

		public void RemoveCity(City city) {
			// This function removes a city from the city list.
			int index = cities.IndexOf(city);
			if (index != -1) {  // found city
				cities.RemoveAt(index);
				Initialize();
			}
		}

		public void CalculateDistances() {
			// This function calculates the distances in the city map.
			iMaxDistance = 0;
			iDistances = new int[cities.Count, cities.Count];
			for (int i = 0; i < cities.Count; i++) {
				for (int j = 0; j < cities.Count; j++) {
					iDistances[i, j] = (int) Math.Sqrt(Math.Pow(((City) cities[i]).x - ((City) cities[j]).x, 2) + Math.Pow(((City) cities[i]).y - ((City) cities[j]).y, 2));
					iMaxDistance = Math.Max(iMaxDistance, iDistances[i, j]);
				}
			}
		}

		#endregion

	}

	public class Entity {

		// This class represents an "entity" - a partial or full solution - that can be added to the pile based on its size and fitness. The actual class is really just a data structure.

		#region Variables and initialization code

		public Signature signature;
		public string path;
		public ArrayList components;
		public ArrayList references;
		public int length;
		public decimal fitness;
		public decimal distance;

		public Entity(int size) {
			signature = new Signature(size);
			path = "";
			fitness = 0;
			length = 0;
			components = new ArrayList();
			references = new ArrayList();
		}

		#endregion

	}

	public class City {

		// This class is a simple data structure for a city.

		#region Variables and initialization code

		internal double x, y;
		internal string name;

		public City(int x, int y, string name) {
			this.x = x;
			this.y = y;
			this.name = name;
		}

		#endregion

	}

	public class Signature {

		// This class is simply a bitmap, representing the list of cities previously visited. It is represented as an array of ints - XORing the integer arrays of two signatures is a very quick way of determining equivalence.

		#region Variables and initialization code

		public int[] signature;
		public char[] endpoints;

		public Signature(int size) {
			signature = new int[(size / 32) + 1];
			endpoints = new char[2];
		}

		#endregion

		#region Helper functions

		public void Add(char city) {
			// Adds a city to the signature list.
			signature[city / 32] |= (1 << (city % 32));
		}

		public void Add(Signature source) {
			// Adds a signature to the signature list.
			for (int i = 0; i < signature.Length; i++)
				signature[i] |= source.signature[i];
		}

		public bool CheckOverlap(char city) {
			// Test whether this city is in the signature.
			return ((signature[city / 32] & (1 << (city % 32))) != 0);
		}

		public bool CheckOverlap(Signature source) {
			// Test whether all cities in the source signature are in the current signature.
			for (int i = 0; i < signature.Length; i++) {
				if ((signature[i] & source.signature[i]) != 0)
					return true;
			}
			return false;
		}

		public bool CheckEquivalence(Signature source) {
			// XOR this signature and the source signature to check for equivalence.
			if ((endpoints[0] != source.endpoints[0]) || (endpoints[1] != source.endpoints[1]))
				return false;
			for (int i = 0; i < signature.Length; i++) {
				if ((signature[i] ^ source.signature[i]) != 0)
					return false;
			}
			return true;
		}

		#endregion

	}
}
